@@ -59,12 +59,12 @@ class BaseEventHooks(object):
     def emit(self, event_name, **kwargs):
         return []
 
-    def register(self, event_name, handler):
+    def register(self, event_name, handler, unique_id=None):
         self._verify_is_callable(handler)
         self._verify_accept_kwargs(handler)
-        self._register(event_name, handler)
+        self._register(event_name, handler, unique_id)
 
-    def unregister(self, event_name, handler):
+    def unregister(self, event_name, handler=None, unique_id=None):
         pass
 
     def _verify_is_callable(self, func):
@@ -117,10 +117,10 @@ class EventHooks(BaseEventHooks):
             responses.append((handler, response))
         return responses
 
-    def _register(self, event_name, handler):
+    def _register(self, event_name, handler, unique_id=None):
         self._handlers[event_name].append(handler)
 
-    def unregister(self, event_name, handler):
+    def unregister(self, event_name, handler, unique_id=None):
         try:
             self._handlers[event_name].remove(handler)
         except ValueError:
@@ -134,35 +134,57 @@ class HierarchicalEmitter(BaseEventHooks):
         # A cache of event name to handler list.
         self._lookup_cache = {}
         self._handlers = _PrefixTrie()
+        # This is used to ensure that unique_id's are only
+        # registered once.
+        self._unique_id_cache = {}
 
     def emit(self, event_name, **kwargs):
         responses = []
         # Invoke the event handlers from most specific
         # to least specific, each time stripping off a dot.
-        logger.debug('emit: %s' % event_name)
-        if event_name in self._lookup_cache:
-            handlers_to_call = self._lookup_cache[event_name]
-        else:
-            handlers_to_call = self._handlers_for_event(event_name)
+        handlers_to_call = self._lookup_cache.get(event_name)
+        if handlers_to_call is None:
+            handlers_to_call = self._handlers.prefix_search(event_name)
             self._lookup_cache[event_name] = handlers_to_call
+        elif not handlers_to_call:
+            # Short circuit and return an empty response is we have
+            # no handlers to call.  This is the common case where
+            # for the majority of signals, nothing is listening.
+            return []
         kwargs['event_name'] = event_name
         responses = []
         for handler in handlers_to_call:
-            logger.debug('emit: calling %s' % handler)
+            logger.debug('Event %s: calling handler %s', event_name, handler)
             response = handler(**kwargs)
             responses.append((handler, response))
         return responses
 
-    def _handlers_for_event(self, event):
-        return self._handlers.prefix_search(event)
-
-    def _register(self, event_name, handler):
+    def _register(self, event_name, handler, unique_id=None):
+        if unique_id is not None:
+            if unique_id in self._unique_id_cache:
+                # We've already registered a handler using this unique_id
+                # so we don't need to register it again.
+                return
+            else:
+                # Note that the trie knows nothing about the unique
+                # id.  We track uniqueness in this class via the
+                # _unique_id_cache.
+                self._handlers.append_item(event_name, handler)
+                self._unique_id_cache[unique_id] = handler
+        else:
+            self._handlers.append_item(event_name, handler)
         # Super simple caching strategy for now, if we change the registrations
         # clear the cache.  This has the opportunity for smarter invalidations.
-        self._handlers.append_item(event_name, handler)
         self._lookup_cache = {}
 
-    def unregister(self, event_name, handler):
+    def unregister(self, event_name, handler=None, unique_id=None):
+        if unique_id is not None:
+            try:
+                handler = self._unique_id_cache.pop(unique_id)
+            except KeyError:
+                # There's no handler matching that unique_id so we have
+                # nothing to unregister.
+                return
         try:
             self._handlers.remove_item(event_name, handler)
             self._lookup_cache = {}
@@ -193,7 +215,12 @@ class _PrefixTrie(object):
 
     """
     def __init__(self):
-        self._root = _Node(None, None)
+        # Each dictionary can be though of as a node, where a node
+        # has values associated with the node, and children is a link
+        # to more nodes.  So 'foo.bar' would have a 'foo' node with
+        # a 'bar' node as a child of foo.
+        # {'foo': {'children': {'bar': {...}}}}.
+        self._root = {'chunk': None, 'children': {}, 'values': None}
 
     def append_item(self, key, value):
         """Add an item to a key.
@@ -204,16 +231,16 @@ class _PrefixTrie(object):
         key_parts = key.split('.')
         current = self._root
         for part in key_parts:
-            if part not in current.children:
-                new_child = _Node(part)
-                current.children[part] = new_child
+            if part not in current['children']:
+                new_child = {'chunk': part, 'values': None, 'children': {}}
+                current['children'][part] = new_child
                 current = new_child
             else:
-                current = current.children[part]
-        if current.values is None:
-            current.values = [value]
+                current = current['children'][part]
+        if current['values'] is None:
+            current['values'] = [value]
         else:
-            current.values.append(value)
+            current['values'].append(value)
 
     def prefix_search(self, key):
         """Collect all items that are prefixes of key.
@@ -226,9 +253,35 @@ class _PrefixTrie(object):
         collected = deque()
         key_parts = key.split('.')
         current = self._root
-        self._get_items(current, key_parts, collected, index=0)
+        self._get_items(current, key_parts, collected, 0)
         return collected
 
+    def _get_items(self, starting_node, key_parts, collected, starting_index):
+        stack = [(starting_node, starting_index)]
+        key_parts_len = len(key_parts)
+        # Traverse down the nodes, where at each level we add the
+        # next part from key_parts as well as the wildcard element '*'.
+        # This means for each node we see we potentially add two more
+        # elements to our stack.
+        while stack:
+            current_node, index = stack.pop()
+            if current_node['values']:
+                seq = reversed(current_node['values'])
+                # We're using extendleft because we want
+                # the values associated with the node furthest
+                # from the root to come before nodes closer
+                # to the root.
+                collected.extendleft(seq)
+            if not index == key_parts_len:
+                children = current_node['children']
+                directs = children.get(key_parts[index])
+                wildcard = children.get('*')
+                next_index = index + 1
+                if wildcard is not None:
+                    stack.append((wildcard, next_index))
+                if directs is not None:
+                    stack.append((directs, next_index))
+
     def remove_item(self, key, value):
         """Remove an item associated with a key.
 
@@ -246,39 +299,17 @@ class _PrefixTrie(object):
         if current_node is None:
             return
         elif index < len(key_parts):
-            next_node = current_node.children.get(key_parts[index])
+            next_node = current_node['children'].get(key_parts[index])
             if next_node is not None:
                 self._remove_item(next_node, key_parts, value, index + 1)
                 if index == len(key_parts) - 1:
-                    next_node.values.remove(value)
-                if not next_node.children and not next_node.values:
+                    next_node['values'].remove(value)
+                if not next_node['children'] and not next_node['values']:
                     # Then this is a leaf node with no values so
                     # we can just delete this link from the parent node.
                     # This makes subsequent search faster in the case
                     # where a key does not exist.
-                    del current_node.children[key_parts[index]]
+                    del current_node['children'][key_parts[index]]
             else:
                 raise ValueError(
                     "key is not in trie: %s" % '.'.join(key_parts))
-
-    def _get_items(self, current_node, key_parts, collected, index):
-        if current_node is None:
-            return
-        if current_node.values:
-            seq = reversed(current_node.values)
-            collected.extendleft(seq)
-        if not len(key_parts) == index:
-            self._get_items(current_node.children.get(key_parts[index]),
-                            key_parts, collected, index + 1)
-            self._get_items(current_node.children.get('*'),
-                            key_parts, collected, index + 1)
-
-
-class _Node(object):
-    def __init__(self, chunk, values=None):
-        self.chunk = chunk
-        self.children = {}
-        self.values = values
-
-    def __repr__(self):
-        return '_Node(chunk=%s, values=%s)' % (self.chunk, self.values)
