@@ -37,6 +37,7 @@ import botocore.base
 import botocore.service
 from botocore.exceptions import ConfigNotFound, EventNotFound, ProfileNotFound
 from botocore.hooks import HierarchicalEmitter, first_non_none_response
+from botocore.provider import get_provider
 from botocore import __version__
 from botocore import handlers
 
@@ -45,8 +46,10 @@ AllEvents = {
     'after-call': '.%s.%s',
     'after-parsed': '.%s.%s.%s.%s',
     'before-call': '.%s.%s',
-    'service-created': ''
-    }
+    'service-created': '',
+    'before-auth': '.%s',
+    'needs-retry': '.%s.%s',
+}
 """
 A dictionary where each key is an event name and the value
 is the formatting string used to construct a new event.
@@ -145,11 +148,18 @@ class Session(object):
         self._config = None
         self._credentials = None
         self._profile_map = None
+        self._provider = None
 
     def _register_builtin_handlers(self, events):
         for event_name, handler in handlers.BUILTIN_HANDLERS:
             self.register(event_name, handler)
 
+    @property
+    def provider(self):
+        if self._provider is None:
+            self._provider = get_provider(self, self.get_variable('provider'))
+        return self._provider
+
     @property
     def available_profiles(self):
         return list(self._build_profile_map().keys())
@@ -182,6 +192,9 @@ class Session(object):
 
     @profile.setter
     def profile(self, profile):
+        # Since provider can be specified in profile, changing the
+        # profile should reset the provider.
+        self._provider = None
         self._profile = profile
 
     def get_variable(self, logical_name, methods=('env', 'config')):
@@ -207,6 +220,7 @@ class Session(object):
         :returns: str value of variable of None if not defined.
         """
         value = None
+        default = None
         if logical_name in self.env_vars:
             config_name, envvar_name, default = self.env_vars[logical_name]
             if logical_name in ('config_file', 'profile'):
@@ -349,46 +363,42 @@ class Session(object):
         """
         return botocore.base.get_data(self, data_path)
 
-    def get_service_data(self, service_name, provider_name='aws'):
+    def get_service_data(self, service_name):
         """
         Retrieve the fully merged data associated with a service.
         """
-        data_path = '%s/%s' % (provider_name, service_name)
+        data_path = '%s/%s' % (self.provider.name, service_name)
         service_data = self.get_data(data_path)
         return service_data
 
-    def get_available_services(self, provider_name='aws'):
+    def get_available_services(self):
         """
         Return a list of names of available services.
         """
-        data_path = '%s' % provider_name
+        data_path = '%s' % self.provider.name
         return self.get_data(data_path)
 
-    def get_service(self, service_name, provider_name='aws'):
+    def get_service(self, service_name):
         """
         Get information about a service.
 
         :type service_name: str
         :param service_name: The name of the service (e.g. 'ec2')
 
-        :type provider_name: str
-        :param provider_name: The name of the provider.  Defaults
-            to 'aws'.
-
         :returns: :class:`botocore.service.Service`
         """
         service = botocore.service.get_service(self, service_name,
-                                               provider_name)
+                                               self.provider)
         event = self.create_event('service-created')
         self._events.emit(event, service=service)
         return service
 
-    def set_debug_logger(self):
+    def set_debug_logger(self, logger_name='botocore'):
         """
         Convenience function to quickly configure full debug output
         to go to the console.
         """
-        log = logging.getLogger('botocore')
+        log = logging.getLogger(logger_name)
         log.setLevel(logging.DEBUG)
 
         # create console handler and set level to debug
@@ -404,7 +414,7 @@ class Session(object):
         # add ch to logger
         log.addHandler(ch)
 
-    def set_file_logger(self, log_level, path):
+    def set_file_logger(self, log_level, path, logger_name='botocore'):
         """
         Convenience function to quickly configure any level of logging
         to a file.
@@ -416,7 +426,7 @@ class Session(object):
         :param path: Path to the log file.  The file will be created
             if it doesn't already exist.
         """
-        log = logging.getLogger('botocore')
+        log = logging.getLogger(logger_name)
         log.setLevel(log_level)
 
         # create console handler and set level to debug
@@ -432,7 +442,7 @@ class Session(object):
         # add ch to logger
         log.addHandler(ch)
 
-    def register(self, event_name, handler):
+    def register(self, event_name, handler, unique_id=None):
         """Register a handler with an event.
 
         :type event_name: str
@@ -444,10 +454,17 @@ class Session(object):
             accept ``**kwargs``.  If either of these preconditions are
             not met, a ``ValueError`` will be raised.
 
-        """
-        self._events.register(event_name, handler)
+        :type unique_id: str
+        :param unique_id: An optional identifier to associate with the
+            registration.  A unique_id can only be used once for
+            the entire session registration (unless it is unregistered).
+            This can be used to prevent an event handler from being
+            registered twice.
 
-    def unregister(self, event_name, handler):
+        """
+        self._events.register(event_name, handler, unique_id)
+
+    def unregister(self, event_name, handler=None, unique_id=None):
         """Unregister a handler with an event.
 
         :type event_name: str
@@ -456,8 +473,28 @@ class Session(object):
         :type handler: callable
         :param handler: The callback to unregister.
 
+        :type unique_id: str
+        :param unique_id: A unique identifier identifying the callback
+            to unregister.  You can provide either the handler or the
+            unique_id, you do not have to provide both.
+
         """
-        self._events.unregister(event_name, handler)
+        self._events.unregister(event_name, handler=handler,
+                                unique_id=unique_id)
+
+    def register_event(self, event_name, fmtstr):
+        """
+        Register a new event.  The event will be added to ``AllEvents``
+        and will then be able to be created using ``create_event``.
+
+        :type event_name: str
+        :param event_name: The base name of the event.
+
+        :type fmtstr: str
+        :param fmtstr: The formatting string for the event.
+        """
+        if event_name not in AllEvents:
+            AllEvents[event_name] = fmtstr
 
     def create_event(self, event_name, *fmtargs):
         """
